import time, types, inspect
from textwrap import dedent

from zope.interface import implements

from epsilon.structlike import record

from twisted.python import log
from twisted.python.components import registerAdapter

from kali.ikali import ICommand
from kali.errors import UsageError, InvalidCommand


class CommandLookupMixin(object):
    implements(ICommand)

    name = None
    usage = 'No usage information'
    help = """No additional help."""

    def locateCommand(self, params):
        """
        Find the handler responsible for a given command.

        The first parameter is popped from C{params} and is treated as the
        command handler name. Handlers should be attributes with the form
        C{cmd_COMMAND}.  If no handler is found for C{cmd} then L{UsageError}
        is raised.

        @param params: The parameters, including the command, for a command
        @type params: C{list}

        @return: The command handler for C{cmd} and the remaining parameters
        @rtype: C{(ICommand, list)}
        """
        cmd = params.pop(0)
        method = getattr(self,
                         'cmd_%s' % cmd.lower(),
                         None)
        if method is None:
            raise InvalidCommand(cmd.lower())

        return ICommand(method), params


class SubCommand(CommandLookupMixin):
    def invoke(self, pseudo):
        raise UsageError('Too few parameters')


class MethodCommand(object):
    implements(ICommand)

    defaultUsage = 'No usage information'
    defaultHelp = """No additional help."""

    def __init__(self, method):
        super(MethodCommand, self).__init__()

        # These attributes may exist but be set to None
        usage = getattr(method, 'usage', None)
        if usage is None:
            usage = self.defaultUsage

        help = getattr(method, 'help', None)
        if help is None:
            help = self.defaultHelp

        lines = dedent(help).splitlines()
        if not lines[0]:
            lines.pop(0)

        self.method = method
        self.params = []
        self.name = method.__name__[len('cmd_'):]
        self.usage = usage
        self.shortHelp = lines[0]
        self.help = '\n'.join(lines)

        self.minargs, self.maxargs = self.getArgLimits()

    def getArgLimits(self):
        minargs, maxargs = arglimits = getattr(self.method, 'arglimits', (None, None))

        args, vararg, varkw, defaults = inspect.getargspec(self.method)
        # Exclude self and pseudo parameters.
        normArgCount = len(args) - 2

        if minargs is None:
            minargs = normArgCount

        if maxargs is None:
            if vararg is None:
                maxargs = normArgCount
            else:
                maxargs = None

        return minargs, maxargs

    def locateCommand(self, params):
        self.params = params
        return self, []

    def invoke(self, pseudo):
        numargs = len(self.params)

        if numargs < self.minargs:
            raise UsageError('Not enough arguments')
        if self.maxargs is not None and numargs > self.maxargs:
            raise UsageError('Too many arguments')

        return self.method(pseudo, *self.params)

registerAdapter(MethodCommand, types.MethodType, ICommand)


class EventHandler(object):
    """
    An event dispatch mechanism.
    """
    def locateHandler(self, event):
        """
        Locate a handler for the specified method.

        Handlers should be named according to the pattern C{event_EVENTNAME}
        and arguments are passed as positional parameters.  Any errors raised
        by event handlers are swallowed and logged.  If no handler is present,
        the event is ignored.

        An example event handler::

            def event_privmsg(self, source, dest, msg): ...

        @returns: a callable handler, or C{None}
        """
        return getattr(self, 'event_%s' % (event,), None)


class PseudoAvatar(CommandLookupMixin):
    """
    Abstract base class for L{Pseudo} avatars.

    C{PseudoAvatar}s allow pseudoclients to customize behaviour on a per-user
    basis.

    @ivar pseudo: The pseudoclient this avatar was created by
    @type pseudo: L{Pseudo} instance

    @ivar nick: The nickname of the user this avatar belongs to
    @type nick: C{str}
    """
    def __init__(self, pseudo, nick):
        """
        @param pseudo: The pseudoclient this avatar was created by
        @type pseudo: L{Pseudo} instance

        @param nick: The nickname of the user this avatar belongs to
        @type nick: C{str}
        """
        self.pseudo = pseudo
        self.nick = nick

    def notice(self, msg):
        """
        Send a notice from this avatar's user via the pseudoclient.
        """
        self.pseudo.notice(self.nick, msg)

    def kill(self, msg):
        """
        Send a KILL message for this user via the pseudoclient.
        """
        self.pseudo.kill(self.nick, msg)


class Pseudo(EventHandler):
    """
    Abstract base class for pseudoclients.

    Anonymous avatars are produced by calling C{anonymousAvatarFactory},
    which should return a L{PseudoAvatar} instance.

    @ivar ctrl: Controller
    @type ctrl: L{ServicesController} instance

    @ivar avatars: Mapping of nicknames to avatars
    @type avatars: C{dict} mapping {str} to L{PseudoAvatar} instances

    @ivar protocol: The protocol that was specified in the connected event
    @type protocol: L{IRCServices} instance
    """
    modes = ''
    anonymousAvatarFactory = PseudoAvatar

    def __init__(self, controller):
        """
        @param controller: Controller
        @type controller: L{ServicesController} instance
        """
        self.ctrl = controller
        self.store = self.ctrl.store
        self.avatars = {}

    def event_connected(self, protocol):
        self.protocol = protocol
        self.protocol.newServiceNick(self.nick, self.name, self.modes)
        self.signon = time.time()

    def event_idle(self, source, dest):
        if dest.lower() == self.nick.lower():
            self.protocol.sendMessage('IDLE', source, '%ld' % self.signon, '0', prefix=self.nick)

    def event_privmsg(self, source, dest, msg):
        if dest.lower() == self.nick.lower():
            self.command(source, msg)

    def locateCommand(self, avatar, params):
        cmd = avatar
        while params:
            cmd, params = cmd.locateCommand(params)

        return cmd

    def command(self, source, msg):
        """
        Dispatch a command to a user's avatar.

        C{cmd} is located on the avatar for C{source}, or an anonmous avatar
        if C{source} is not found in L{avatars}, and invoked. If L{UsageError}
        is raised by the handler the error and usage information is displayed,
        otherwise the error is swallowed and logged.
        """
        params = msg.split()
        if params:
            avatar = self.avatars.get(source)
            if avatar is None:
                avatar = self.avatars[source] = self.anonymousAvatarFactory(self, source)

            try:
                try:
                    cmd = self.locateCommand(avatar, params)
                    return cmd.invoke(self)
                except UsageError, (e,):
                    raise UsageError(cmd, e)
            except InvalidCommand, (e,):
                avatar.notice('Unknown command: ' + e)
            except UsageError, (cmd, e):
                avatar.notice('Invalid usage: ' + e)
                # XXX: PseudoAvatar doesn't have a usage attribute.
                if getattr(cmd, 'usage', None):
                    avatar.notice('Usage: %s' % (cmd.usage,))
            except:
                log.err()

    def notice(self, nick, msg):
        """
        Send a notice from C{nick} via the protocol.
        """
        self.protocol.notice(self.nick, nick, msg)

    def kill(self, nick, msg):
        """
        Send a KILL message for C{nick} via the protocol.
        """
        self.protocol.kill(nick, msg, source=self.nick)


class _ModesMixin(object):
    """
    Mixin for IRC objects with modes.
    """
    def hasMode(self, mode):
        """
        Determines whether a particular mode is set.

        @param mode: A single character representing the mode to check
        @type mode: C{str}

        @rtype: C{bool}
        """
        return mode in self.modes

    def addModes(self, modes):
        """
        Add modes from the set of currently active modes.

        @param modes: A string of mode characters to be removed.
        @type mode: C{str}
        """
        self.modes.update(set(modes))

    def removeModes(self, modes):
        """
        Remove modes from the set of currently active modes.

        @param modes: A string of mode characters to be removed.
        @type mode: C{str}
        """
        self.modes.difference_update(set(modes))


class User(record('nick user host name modes signon server'), _ModesMixin):
    """
    Represents the state of a single IRC user.
    """
    def __init__(self, *a, **kw):
        super(User, self).__init__(*a, **kw)
        self.registered = False
        self.channels = set()


class Channel(record('name'), _ModesMixin):
    """
    Represents the state of an IRC channel.
    """
    def __init__(self, *a, **kw):
        super(Channel, self).__init__(*a, **kw)
        self.modes = set()
        self.users = set()


class ServicesController(EventHandler):
    def __init__(self, store, params):
        self.store = store
        self.params = params
        self.handlers = [self]

        self.servers = {}
        self.users = {}
        self.channels = {}

    def addService(self, service):
        self.handlers.append(service)

    def event(self, event, *params):
        for p in self.handlers:
            self.sendEvent(p, event, *params)

    def sendEvent(self, receiver, event, *params):
        """
        Dispatch an event to a handler, if it exists.
        """
        method = receiver.locateHandler(event)
        if method is None:
            return

        try:
            method(*params)
        except:
            log.err()

    def getUser(self, nick):
        return self.users.get(nick.lower())

    def getChannel(self, channelName):
        return self.channels[channelName.lower()]

    def createUser(self, user):
        nick = user['nick']
        u = self.users[nick.lower()] = User(
            nick=nick,
            user=user['username'],
            host=user['hostname'],
            name=user['realname'],
            modes=set(user['modes']),
            signon=user['timestamp'],
            server=user['server'])

        self.event('newuser', u)

    def event_connected(self, protocol):
        self.protocol = protocol

    def event_nickchange(self, old, new):
        user = self.users.pop(old.lower())
        user.nick = new
        self.users[new.lower()] = user

    def event_quit(self, nick, msg):
        user = self.getUser(nick)
        for channel in user.channels.copy():
            self.event('left', channel, user)
        self.users.pop(nick.lower())

    def event_opertype(self, nick, opertype):
        user = self.users[nick.lower()]
        user.opertype = opertype

    def event_newserver(self, server, location):
        self.servers[server] = location

    def event_modeadded(self, source, target, targetKind, (mode, param)):
        if targetKind == 'user':
            user = self.getUser(target)
            if param is not None:
                raise NotImplementedError('%r: %r' % (mode, param))
            else:
                user.addModes(mode)
        elif targetKind == 'channel':
            channel = self.getChannel(target)
            if param is not None:
                raise NotImplementedError('%r: %r' % (mode, param))
            else:
                channel.addModes(mode)
        else:
            raise NotImplementedError(targetKind)

    def event_moderemoved(self, source, target, targetKind, (mode, param)):
        if targetKind == 'user':
            user = self.getUser(target)
            if param is not None:
                raise NotImplementedError('%r: %r' % (mode, param))
            else:
                user.removeModes(mode)
        elif targetKind == 'channel':
            channel = self.getChannel(target)
            if param is not None:
                raise NotImplementedError('%r: %r' % (mode, param))
            else:
                channel.removeModes(mode)
        else:
            raise NotImplementedError(targetKind)

    def event_joined(self, channelName, nick):
        try:
            channel = self.getChannel(channelName)
        except KeyError:
            channel = self.channels[channelName] = Channel(channelName)

        user = self.getUser(nick)
        channel.users.add(user)
        user.channels.add(channel)

    def event_parted(self, channelName, nick, message):
        user = self.getUser(nick)
        channel = self.getChannel(channelName)
        self.event('left', channel, user)

    def event_left(self, channel, user):
        channel.users.remove(user)
        user.channels.remove(channel)
        if len(channel.users) == 0:
            del self.channels[channel.name]
