from twisted.internet import reactor

from axiom.item import Item, normalize
from axiom.attributes import text

from kali.controller import Pseudo, PseudoAvatar
from kali.util import usage, HelpMixin


class NickServError(Exception):
    pass


class Anonymous(PseudoAvatar, HelpMixin):
    def __init__(self, pseudo, nick):
        super(Anonymous, self).__init__(pseudo, nick)
        self.identifyAttempts = pseudo.maxPasswordAttempts

    @usage('IDENTIFY <password>')
    def cmd_identify(self, pseudo, password):
        """
        Identifies you as the owner of your current nickname.

        Usage: IDENTIFY <password>

        Identifying yourself as the owner of a nickname grants you permission
        to use other nickname dependent services. You should use the same
        <password> that you used for the REGISTER command.
        """
        try:
            accepted = pseudo.identify(self.nick, password)
            if accepted:
                self.notice('Password accepted, you are now recognized.')
            else:
                self.notice('Password incorrect.')
                self.identifyAttempts -= 1
                if self.identifyAttempts == 1:
                    self.notice('WARNING: One more incorrect password attempt and you will be disconnected.')
                elif self.identifyAttempts == 0:
                    self.kill('Too many invalid passwords (%s)' % (self.nick,))
        except NickServError, e:
            self.notice(e)

    @usage('REGISTER <password>')
    def cmd_register(self, pseudo, password):
        """
        Registers the current nickname.

        Usage: REGISTER <password>

        Registering a nickname grants you permission to use that nickname
        to access other nickname dependent services. You should use a
        password other people are unlikely to guess.
        """
        try:
            accepted = pseudo.register(self.nick, password)
            if accepted:
                self.notice('Registration successful, you are now recognised.')
        except NickServError, e:
            self.notice(e)

    @usage('STATUS <nickname>')
    def cmd_status(self, pseudo, nickname):
        """
        Query the status of the given nickname. The response has the following format:

            STATUS <nickname> <status code>

        where <status code> is one of the following:

            0 - no such user online or nickname not registered
            1 - user not recognized as nickname's owner
            2 - user recognized as owner via access list only
            3 - user recognized as owner via password identification
        """
        self.notice('STATUS %s %s' % (nickname, self.getStatus(nickname)))

    def getStatus(self, nickname):
        """
        Get the applicable status code for a nickname.
        """
        user = self.pseudo.ctrl.getUser(nickname)
        if user is None:
            return 0

        nick = self.pseudo.getNick(nickname)
        if nick is None:
            return 0

        if not user.registered:
            return 1

        return 3


class Authenticated(Anonymous):
    help = """
    NickServ provides nickname registration services, allowing you to reserve
    control over a particular nickname.
    """


class TierThree(Authenticated):
    pass


class Nick(Item):
    typeName = normalize('kali.controller.Nick')

    # We should probably have a uniqueness constraint on nick, but Axiom
    # doesn't do that: http://divmod.org/trac/ticket/366
    nick = text(allowNone=False)
    password = text(allowNone=False)
    kill = text(default=u'off')


class NickServ(Pseudo):
    nick = 'NickServ'
    name = 'Nickname Services'
    anonymousAvatarFactory = Anonymous

    maxPasswordAttempts = 5

    killDelays = {
        u'off': None,
        u'on': 60,
        u'quick': 20,
        }

    tierAvatars = {
        'tier-3': TierThree,
        }

    def __init__(self, *a, **kw):
        super(NickServ, self).__init__(*a, **kw)

        self.collisions = {}
        self.opertypes = {}

    def createAvatar(self, nick):
        """
        Construct an authenticated avatar for L{nick}.
        """
        opertype = self.opertypes.get(nick)
        if opertype is not None:
            avatarCls = self.tierAvatars.get(opertype)
            if avatarCls is not None:
                return avatarCls(self, nick)

        return Authenticated(self, nick)

    def login(self, nick):
        avatar = self.createAvatar(nick)
        self.avatars[nick] = avatar
        self.cancelCollision(nick)
        self.ctrl.getUser(nick).registered = True
        # XXX: set mode +r

    def logout(self, nick):
        self.ctrl.getUser(nick).registered = False
        # XXX: set mode -r

        self.cancelCollision(nick)
        self.avatars.pop(nick, None)
        self.opertypes.pop(nick, None)

    def scheduleCollision(self, nick, delay):
        """
        Schedule a nick kill.

        @param nick: The nickname to kill.
        @param delay: Time until kill, in seconds.
        """
        self.cancelCollision(nick)
        self.collisions[nick] = reactor.callLater(delay, self.collide, nick)

    def cancelCollision(self, nick):
        """
        Cancel a pending nick kill.

        @param nick: The nickname to cancel a kill for.
        """
        col = self.collisions.pop(nick, None)
        if col is not None:
            col.cancel()

    def event_newuser(self, user):
        nick = user.nick
        self.validate(nick)

    def event_nickchange(self, old, new):
        # By the time we get here, the user has already been renamed.
        self.logout(new)
        self.validate(new)

    def event_quit(self, nick, msg):
        # By the time we get here, the user is already gone.
        pass

    def event_opertype(self, nick, opertype):
        self.opertypes[nick] = opertype
        if opertype is not None:
            self.login(nick)
        else:
            self.logout(nick)

    def getNick(self, nickname):
        if isinstance(nickname, str):
            nickname = unicode(nickname, 'ascii')
        return self.store.findUnique(Nick, Nick.nick == nickname, default=None)

    def setNick(self, nickname, password, **kw):
        if isinstance(nickname, str):
            nickname = unicode(nickname, 'ascii')
        if isinstance(password, str):
            password = unicode(password, 'ascii')
        # Since we don't get uniqueness, we use Axiom's lack of concurrency
        # to our advantage here.
        if self.store.findFirst(Nick, Nick.nick == nickname, default=None):
            raise NickServError('This nickname is already registered.')
        return Nick(store=self.store, nick=nickname, password=password, **kw)

    def validate(self, nick):
        """
        Check to see if a new nickname needs to identify.

        A nick kill is scheduled if appropriate.

        @param nick: The nickname to check.
        """
        nick = nick.lower()
        n = self.getNick(nick)
        if n is not None:
            self.notice(nick, 'This nickname is registered. If this is your nickname type: /msg NickServ IDENTIFY <password>. Otherwise, please choose a different nickname.')

            delay = self.killDelays.get(n.kill)
            if delay is not None:
                self.scheduleCollision(nick, delay)

    def collide(self, nick):
        self.collisions.pop(nick, None)
        self.protocol.kill(nick, 'Nick kill enforced', self.nick)

    def identify(self, nick, password):
        """
        Attempt authentication for a nickname.

        @param nick: The nickname being authenticated for.
        @param password: The password specified.
        """
        n = self.getNick(nick)
        if n is None:
            raise NickServError('This nickname is not registered.')

        if n.password == password:
            self.login(nick)
            return True

        return False

    def register(self, nick, password, **kw):
        """
        Attempt to register a nickname.

        @param nick: The nickname being registration.
        @param password: The password for the nickname.
        """
        n = self.setNick(nick, password)
        self.login(nick)
        return True
