import fnmatch, re

from axiom.item import Item
from axiom.attributes import text, integer

from kali.controller import Pseudo, PseudoAvatar, SubCommand
from kali.util import HelpMixin, usage, tabulate
from kali.errors import UsageError


class Anonymous(PseudoAvatar):
    def cmd_help(self, pseudo, *params):
        # XXX: use HelpMixin?
        self.notice('You have no access to OperServ.')


class TierOne(PseudoAvatar, HelpMixin):
    help = """
    OperServ provides various IRC operator related network services.
    """

    @usage('GLOBAL <message>')
    def cmd_global(self, pseudo, *params):
        """
        Send a message to all users.

        Usage: GLOBAL <message>

        Sends a notice, with the text <message>, to all users on the network.
        """
        msg = ' '.join(params)
        msg = '(notice from %s) %s' % (self.nick, msg)
        self.pseudo.messenger.notice('$*', msg)


class TierTwo(TierOne):
    """
    Avatar for tier-2 opers.
    """


class Session(SubCommand):
    name = 'session'

    @usage('SESSION LIST <threshold>')
    def cmd_list(self, pseudo, threshold):
        """
        Lists host sessions.

        Usage: SESSION LIST <threshold>

        Lists hosts with at least <threshold> sessions. The threshold must be
        a number greater than 1, to prevent accidentally listing countless
        single session hosts.
        """
        threshold = int(threshold)
        if threshold <= 1:
            raise UsageError('The threshold must be greater than 1')

        pseudo.notice('Hosts with at least %d sessions:' % (threshold,))

        data = [(sessions, host) for sessions, host in pseudo.sessions.iteritems()
                if sessions > threshold]
        data.sort(reverse=True)

        for row in tabulate(['Sessions', 'Host'], data):
            pseudo.notice(row)

    @usage('SESSION VIEW <host>')
    def cmd_view(self, pseudo, host):
        """
        View host's detailed session information.

        Usage: SESSION VIEW <host>

        Displays detailed information about a specific host, including the
        current session count and limit. Any wildcards present in <host> are
        ignored.
        """
        data = pseudo.sessions.get(host, None)
        if data is not None:
            sessions, host = data
            limitException = pseudo.limitExceptionForHost(host)
            if limitException is not None:
                sessionLimit = limitException.limit
            else:
                sessionLimit = pseudo.defaultSessionLimit
            msg = 'The host \037%s\037 currently has \002%d\002 session(s) with a limit of \002%d\002.' % (host, sessions, sessionLimit)
        else:
            msg = 'No session entries for host \037%s\037.' % (host,)

        pseudo.notice(msg)


class TierThree(TierTwo):
    """
    Avatar for tier-3 opers.
    """
    cmd_session = Session()


class Messenger(Pseudo):
    nick = 'Shadowfire'
    name = 'Network Messaging Service'

    def command(self, source, cmd):
        pass


class LimitException(Item):
    typeName = 'kali_oper_limit_exception'
    schemaVersion = 1

    exceptionID = integer(allowNone=False)
    expiry = text()
    mask = text(allowNone=False)
    reason = text(allowNone=False)
    limit = integer(allowNone=False)


class OperServ(Pseudo):
    nick = 'OperServ'
    name = 'Oper Network Services'
    anonymousAvatarFactory = Anonymous

    # XXX: defaultSessionLimit should be configurable
    defaultSessionLimit = 5

    tierAvatars = {
        'tier-1': TierOne,
        'tier-2': TierTwo,
        'tier-3': TierThree}

    def __init__(self, controller):
        super(OperServ, self).__init__(controller)
        self.messenger = Messenger(controller)
        self.ctrl.addService(self.messenger)

        self.sessions = {}

    def allLimitExceptions(self):
        return self.store.query(LimitException)

    def limitExceptionForHost(self, host):
        for le in self.allLimitExceptions():
            match = re.match(fnmatch.translate(le.mask), host)
            if match:
                return le

        return None

    def getAvatar(self, nick, opertype):
        avatarCls = self.tierAvatars.get(opertype)
        if avatarCls is not None:
            return avatarCls(self, nick)

        return None

    def login(self, nick, opertype):
        avatar = self.getAvatar(nick, opertype)
        if avatar is not None:
            self.avatars[nick] = avatar

    def logout(self, nick):
        self.avatars.pop(nick, None)

    def event_newuser(self, user):
        sessions = self.sessions
        sessions[user.host] = sessions.setdefault(user.host, 0) + 1

    def event_opertype(self, nick, opertype):
        if opertype is not None:
            self.login(nick, opertype)
        else:
            self.logout(nick)

    def event_nickchange(self, old, new):
        avatar = self.avatars.pop(old, None)
        if avatar is not None:
            self.avatars[new] = avatar

    def event_quit(self, nick, msg):
        self.logout(nick)
