from twisted.trial.unittest import TestCase

from kali.controller import (Pseudo, PseudoAvatar, SubCommand,
    ServicesController, User, Channel, MethodCommand)
from kali.errors import UsageError, InvalidCommand
from kali.util import usage


class TestingController(object):
    def __init__(self):
        self.store = None


class TestingSubCommand(SubCommand):
    name = 'sub'

    def cmd_known(self, pseudo):
        return 301
    cmd_known.id = 300


class TestingAvatar(PseudoAvatar):
    def cmd_known(self, pseudo):
        return 201
    cmd_known.id = 200

    cmd_sub = TestingSubCommand()
    cmd_sub.id = 202

    def cmd_varargs(self, pseudo, a, b, *c):
        return 203

    def cmd_novarargs(self, pseudo, a, b, c):
        return 204

    @usage('LIMITS foo [bar] [baz]', min=1, max=3)
    def cmd_limits(self, pseudo, a, *bc):
        return 205

    def cmd_maybebrokenname(self, pseudo):
        pass


class TestingClient(Pseudo):
    anonymousAvatarFactory = TestingAvatar

    def event_known(self):
        return 101


class EventHandlingTests(TestCase):
    def setUp(self):
        self.ctrl = TestingController()
        self.client = TestingClient(self.ctrl)

    def test_locateUnknownEvent(self):
        """
        Locating an unknown event handler.
        """
        handler = self.client.locateHandler('doesnotexist')
        self.assertIdentical(handler, None)

    def test_locateKnownEvent(self):
        """
        Locating a known event handler.
        """
        handler = self.client.locateHandler('known')
        self.assertEqual(handler(), 101)


class CommandHandlingTests(TestCase):
    def setUp(self):
        self.ctrl = TestingController()
        self.pseudo = TestingClient(self.ctrl)
        self.avatar = TestingAvatar(self.pseudo, None)

    def locateCommand(self, params):
        return self.pseudo.locateCommand(self.avatar, params)

    def test_locateUnknownCommand(self):
        """
        Locating an unknown command.
        """
        params = 'doesnotexist'.split()
        self.assertRaises(InvalidCommand, self.locateCommand, params)

    def test_locateTopLevelCommand(self):
        """
        Locating a top-level command.
        """
        params = 'known'.split()
        cmd = self.locateCommand(params)
        self.assertEqual(cmd.method.id, 200)

    def test_invokeTopLevelCommand(self):
        """
        Invoking a top-level command.
        """
        source = 'source'
        msg = 'known'
        self.assertEqual(self.pseudo.command(source, msg), 201)

    def test_locateSubCommand(self):
        """
        Locating a sub-command.
        """
        params = 'sub known'.split()
        cmd = self.locateCommand(params)
        self.assertEqual(cmd.method.id, 300)

    def test_invokeSubCommand(self):
        """
        Invoking a sub-command.
        """
        source = 'source'
        msg = 'sub known'
        self.assertEqual(self.pseudo.command(source, msg), 301)

    def test_tooFewVarArgs(self):
        params = 'varargs'.split()
        cmd = self.locateCommand(params)
        self.assertRaises(UsageError, cmd.invoke, self.pseudo)

    def test_varargs(self):
        params = 'varargs a b c d e f'.split()
        cmd = self.locateCommand(params)
        self.assertEqual(cmd.invoke(self.pseudo), 203)

    def test_tooFewArgs(self):
        params = 'novarargs a'.split()
        cmd = self.locateCommand(params)
        self.assertRaises(UsageError, cmd.invoke, self.pseudo)

    def test_tooManyArgs(self):
        params = 'novarargs a b c d'.split()
        cmd = self.locateCommand(params)
        self.assertRaises(UsageError, cmd.invoke, self.pseudo)

    def test_novarargs(self):
        params = 'novarargs a b c'.split()
        cmd = self.locateCommand(params)
        self.assertEqual(cmd.invoke(self.pseudo), 204)

    def test_withinLimits(self):
        params = 'limits a b'.split()
        cmd = self.locateCommand(params)
        self.assertEqual(cmd.invoke(self.pseudo), 205)

    def test_underLimits(self):
        params = 'limits'.split()
        cmd = self.locateCommand(params)
        self.assertRaises(UsageError, cmd.invoke, self.pseudo)

    def test_overLimits(self):
        params = 'limits a b c d e f'.split()
        cmd = self.locateCommand(params)
        self.assertRaises(UsageError, cmd.invoke, self.pseudo)

    def test_maybeBrokenName(self):
        """
        Names are properly formed.

        There used to be code doing name.strip('cmd_') which caused commands
        like "mailer" to turn into "ailer".
        """
        params = 'maybebrokenname'.split()
        cmd = self.locateCommand(params)
        self.assertEqual(cmd.name, 'maybebrokenname')


class EventCollector(object):
    """
    Event handler that just collects events.
    """
    def __init__(self):
        self.events = []

    def locateHandler(self, event):
        """
        Return a handler that just appends to the list of events.
        """
        def _handler(*params):
            self.events.append((event,) + params)
        return _handler


class ControllerTests(TestCase):
    def setUp(self):
        self.services = ServicesController(None, None)
        self.collector = EventCollector()
        self.services.addService(self.collector)

        self.testuser = User('testnick', 'foo', 'bar.com', 'Some User', set(), '1123', 'some.server')
        self.services.users[self.testuser.nick] = self.testuser

        self.testuser2 = User('testnick2', 'foo', 'bar.com', 'Some User', set(), '1123', 'some.server')
        self.services.users[self.testuser2.nick] = self.testuser2

        self.testchan = Channel('#testchan')
        self.services.channels[self.testchan.name] = self.testchan

    def test_userModeChanges(self):
        """
        Test that user modes are updated on mode change events.
        """
        for mode in [('i', None), ('w', None)]:
            self.services.event('modeadded', 'fake.source', self.testuser.nick, 'user', mode)
        self.assertEqual(self.testuser.modes, set('iw'))
        self.services.event('moderemoved', 'fake.source', self.testuser.nick, 'user', ('i', None))
        self.assertEqual(self.testuser.modes, set('w'))

    def test_channelModeChanges(self):
        """
        Test that channel modes are updated on mode change events.
        """
        for mode in [('n', None), ('s', None)]:
            self.services.event('modeadded', 'fake.source', self.testchan.name, 'channel', mode)
        self.assertEqual(self.testchan.modes, set('ns'))
        self.services.event('moderemoved', 'fake.source', self.testchan.name, 'channel', ('s', None))

    def test_getUser(self):
        """
        getUser returns a user object, or C{None}.
        """
        self.assertIdentical(self.services.getUser('testnick'), self.testuser)
        self.assertIdentical(self.services.getUser('doesntexistnick'), None)

    def test_getChannel(self):
        """
        getChannel retrieves a channel object.
        """
        self.assertRaises(KeyError, self.services.getChannel, '#doesntexist')
        self.assertNotIdentical(self.services.getChannel('#testchan'), None)

    def test_joinAndPartChannel(self):
        """
        A channel is created when the first user joins, and destroyed when the
        last users leaves. A channel also keeps a list of users currently
        joined to it.
        """
        self.assertNotIn('#testing', self.services.channels)
        self.services.event('joined', '#testing', self.testuser.nick)
        self.assertIn('#testing', self.services.channels)

        channel = self.services.getChannel('#testing')
        self.assertIn(self.testuser, channel.users)
        self.services.event('joined', '#testing', self.testuser2.nick)
        self.assertIn(self.testuser2, channel.users)

        self.services.event('left', channel, self.testuser)
        self.services.event('left', channel, self.testuser2)
        self.assertNotIn('#testing', self.services.channels)

    def test_joinAndPartUser(self):
        """
        A user keeps a list of the channels they are joined to.
        """
        self.assertNotIn(self.testchan, self.testuser.channels)
        self.services.event('joined', self.testchan.name, self.testuser.nick)
        self.assertIn(self.testchan, self.testuser.channels)
        self.services.event('left', self.testchan, self.testuser)
        self.assertNotIn(self.testchan, self.testuser.channels)

    def test_quitChannel(self):
        """
        Users that quit leave all of the channels they are joined to.
        """
        self.services.event('joined', self.testchan.name, self.testuser.nick)
        self.services.event('quit', self.testuser.nick, 'lulz quit')
        self.assertNotIn(self.testchan, self.testuser.channels)
        self.assertNotIn(self.testuser, self.testchan.users)

    def test_partChannel(self):
        """
        Users that part a channel leave it.
        """
        self.services.event('joined', self.testchan.name, self.testuser.nick)
        self.services.event('parted', self.testchan.name, self.testuser.nick, 'lulz parted')
        self.assertNotIn(self.testchan, self.testuser.channels)
        self.assertNotIn(self.testuser, self.testchan.users)


class MethodCommandTests(TestCase):
    def _m0(self, pseudo):
        pass

    def _m2(self, pseudo, p1, p2):
        pass

    def _m0va(self, pseudo, *args):
        pass

    def _m2va(self, pseudo, p1, p2, *args):
        pass

    def test_argLimits(self):
        """
        Verify argument limits on normal methods.
        """
        m0 = MethodCommand(self._m0)
        self.assertEqual(m0.minargs, 0)
        self.assertEqual(m0.maxargs, 0)

        m2 = MethodCommand(self._m2)
        self.assertEqual(m2.minargs, 2)
        self.assertEqual(m2.maxargs, 2)

    def test_varargLimits(self):
        """
        Verify argument limits on methods with *args.
        """
        m0 = MethodCommand(self._m0va)
        self.assertEqual(m0.minargs, 0)
        self.assertEqual(m0.maxargs, None)

        m2 = MethodCommand(self._m2va)
        self.assertEqual(m2.minargs, 2)
        self.assertEqual(m2.maxargs, None)
