from twisted.trial.unittest import TestCase

from axiom.store import Store
from kali.nickserv import NickServ, Authenticated, TierThree, Nick, NickServError
from kali.controller import ServicesController, User

class FakeProtocol(object):
    """
    Mock object for testing protocol interactions.
    """
    def __init__(self):
        self.stubs = {}

    def addStub(self, stub, *args, **kw):
        """
        Add a stub method that can be called.
        
        XXX: This isn't very flexible. You're limited to a single set
        of params for a particular method. (We can fix this if we ever
        need to.)
        """
        self.stubs[stub] = {'args': args, 'kw': kw}

    def __getattr__(self, attr):
        """
        Return a stub method that asserts the parameters it gets
        matches those specified.
        """
        stub = self.stubs[attr]
        def callStub(*args, **kw):
            assert stub['args'] == args
            assert stub['kw'] == kw
        return callStub


class NickServTests(TestCase):
    def setUp(self):
        self.nickserv = NickServ(ServicesController(Store(), {}))
        self.nickserv.protocol = FakeProtocol()

    def test_collide(self):
        """
        Schedule and cancel a nick collision.
        """
        self.nickserv.scheduleCollision('foo', 30)
        self.assertIn('foo', self.nickserv.collisions)
        self.nickserv.cancelCollision('foo')
        self.assertNotIn('foo', self.nickserv.collisions)

    def test_doubleCollide(self):
        """
        Schedule, reschedule and cancel a nick collision.
        """
        self.nickserv.scheduleCollision('foo', 30)
        self.nickserv.scheduleCollision('foo', 30)
        self.assertIn('foo', self.nickserv.collisions)
        self.nickserv.cancelCollision('foo')
        self.assertNotIn('foo', self.nickserv.collisions)

    def test_noneCollide(self):
        """
        Cancel a non-existant nick collision.
        """
        self.nickserv.cancelCollision('foo')

    def test_collideDoubleCancel(self):
        """
        Cancel a scheduled collision twice.
        """
        self.nickserv.scheduleCollision('foo', 30)
        self.nickserv.cancelCollision('foo')
        self.nickserv.cancelCollision('foo')
        self.assertNotIn('foo', self.nickserv.collisions)

    def test_getAvatarNotOper(self):
        """
        Get an Authenticated avatar for a user.
        """
        avatar = self.nickserv.createAvatar('foo')
        self.assertEqual(type(avatar), Authenticated)

    def test_getAvatarOper(self):
        """
        Get a TierThree avatar for a suitable oper.
        """
        self.nickserv.opertypes['foo'] = 'tier-3'
        avatar = self.nickserv.createAvatar('foo')
        self.assertEqual(type(avatar), TierThree)

    def test_getAvatarWrongOper(self):
        """
        Get an Authenticated avatar for an unknown oper.
        """
        self.nickserv.opertypes['foo'] = 'tier-2'
        avatar = self.nickserv.createAvatar('foo')
        self.assertIdentical(type(avatar), Authenticated)

    def setFakeUser(self, nick):
        """
        Create a fake user that we can test against.
        """
        self.nickserv.ctrl.users[nick] = User(nick=nick,
                                              user='user',
                                              host='host',
                                              name='Foo',
                                              modes=set(),
                                              signon=123456789,
                                              server='irc.mynet.org')

    def test_register(self):
        """
        Register a previously unregistered user.
        """
        self.setFakeUser('foo')
        self.nickserv.protocol.addStub('setAccount', 'foo', 'foo')
        self.nickserv.register('foo', 'pass')
        self.assertTrue(self.nickserv.ctrl.users['foo'].registered)
    
    def test_alreadyRegistered(self):
        """
        Register a previously registered user, which fails.
        """
        self.setFakeUser('foo')
        Nick(store=self.nickserv.store, nick=u'foo', password=u'pass')
        self.assertRaises(NickServError, self.nickserv.register, 'foo', 'pass')
    
