from twisted.trial.unittest import TestCase

from kali.protocol import parsemsg, IRCBadMessage, IRCServices, IRCBadModes, parseModes, InspircdServices


class TestingIRCServices(IRCServices):
    pass


class MessageParsingTests(TestCase):
    def test_emptyLine(self):
        """
        Receiving an empty line is an error.
        """
        self.assertRaises(IRCBadMessage, parsemsg, '')

    def test_simpleMessage(self):
        """
        A message with no prefix, trailer, or arguments.
        """
        self.assertEqual(parsemsg('SIMPLE'),
                         ('', 'SIMPLE', []))

    def test_complexMessage(self):
        """
        A message with a prefix, arguments, and a trailer.
        """
        self.assertEqual(parsemsg(':source TEST some stuff :and more'),
                         ('source', 'TEST', ['some', 'stuff', 'and more']))


class ModeParsingTests(TestCase):
    paramModes = ('klb', 'b')

    def test_emptyModes(self):
        """
        An exception is raised if the mode string is empty.
        """
        self.assertRaises(IRCBadModes, parseModes, '', [])

    def test_malformedModes(self):
        """
        An exception is raised if the mode string does not start with plus or
        minus.
        """
        self.assertRaises(IRCBadModes, parseModes, 'foo', [])
        self.assertRaises(IRCBadModes, parseModes, '%', [])

    def test_nullModes(self):
        """
        An exception is raised if no mode characters are present in the mode
        string.
        """
        self.assertRaises(IRCBadModes, parseModes, '+', [])
        self.assertRaises(IRCBadModes, parseModes, '-', [])

    def test_singleMode(self):
        """
        Parse a single mode setting with no parameters.
        """
        added, removed = parseModes('+s', [])
        self.assertEqual(added, [('s', None)])
        self.assertEqual(removed, [])

        added, removed = parseModes('-s', [])
        self.assertEqual(added, [])
        self.assertEqual(removed, [('s', None)])

    def test_singleDirection(self):
        """
        Parse a single-direction mode setting with multiple modes, and no
        parameters.
        """
        added, removed = parseModes('+stn', [])
        self.assertEqual(added, [('s', None),
                                 ('t', None),
                                 ('n', None)])
        self.assertEqual(removed, [])

        added, removed = parseModes('-nt', [])
        self.assertEqual(added, [])
        self.assertEqual(removed, [('n', None),
                                   ('t', None)])

    def test_multiDirection(self):
        """
        Parse a multi-direction mode setting with no parameters.
        """
        added, removed = parseModes('+s-n+ti', [])
        self.assertEqual(added, [('s', None),
                                 ('t', None),
                                 ('i', None)])
        self.assertEqual(removed, [('n', None)])

    def test_consecutiveDirection(self):
        """
        Parse a multi-direction mode setting containing two consecutive mode
        sequences with the same direction.

        XXX: This should possibly be an error; if that decision is made, this
             test should be changed appropriately.
        """
        added, removed = parseModes('+sn+ti', [])
        self.assertEqual(added, [('s', None),
                                 ('n', None),
                                 ('t', None),
                                 ('i', None)])
        self.assertEqual(removed, [])

    def test_mismatchedParams(self):
        """
        An exception is raised if the number of mode parameters does not match
        the number of modes expecting parameters.
        """
        self.assertRaises(IRCBadModes, parseModes, '+k', [], self.paramModes)
        self.assertRaises(IRCBadModes, parseModes, '+kl', ['foo', '10', 'lulz_extra_param'], self.paramModes)

    def xtest_parameters(self):
        """
        Parse a mode string containing modes with parameters.
        """
        added, removed = parseModes('+klbb', ['somekey', '42', 'nick!user@host', 'other!*@*'], self.paramModes)
        self.assertEqual(added, [('k', 'somekey'),
                                 ('l', '42'),
                                 ('b', 'nick!user@host'),
                                 ('b', 'other!*@*')])
        self.assertEqual(removed, [])

        added, removed = parseModes('-klbb', ['nick!user@host', 'other!*@*'], self.paramModes)
        self.assertEqual(added, [])
        self.assertEqual(removed, [('k', None),
                                   ('l', None),
                                   ('b', 'nick!user@host'),
                                   ('b', 'other!*@*')])


class TestEventCollector(object):
    """
    Fake controller that just collects events.
    """
    def __init__(self, events):
        self.events = events

    def event(self, name, *params):
        """
        Collect events.
        """
        self.events.append((name,) + tuple(params))


class ProtocolTests(TestCase):
    cannedModes = [('modeadded',   'irc.mynet.org', '#opers', 'channel', ('b', '*!user@spy.com')),
                   ('modeadded',   'irc.mynet.org', '#opers', 'channel', ('b', '*!person@watching.us')),
                   ('moderemoved', 'irc.mynet.org', '#opers', 'channel', ('s', None))]
    def setUp(self):
        self.protocol = InspircdServices()
        self.events = []
        self.protocol.services = TestEventCollector(self.events)

    def test_fmode(self):
        """
        Test that FMODE emits the correct events.
        """
        self.protocol.irc_FMODE('irc.mynet.org', '#opers', '115432135', '+bb-s', '*!user@spy.com', '*!person@watching.us')
        self.assertEqual(self.events, self.cannedModes)

    def test_modeLame(self):
        """
        Test that MODE handles a stupidly-formed mode string.
        """
        self.protocol.irc_MODE('irc.mynet.org', '#opers', 'bb', '*!user@spy.com', '*!person@watching.us')
        self.protocol.irc_MODE('irc.mynet.org', '#opers', '-s')
        self.assertEqual(self.events, self.cannedModes)

    def test_mode(self):
        """
        Test that MODE emits the correct events.
        """
        self.protocol.irc_MODE('irc.mynet.org', '#opers', '+bb-s', '*!user@spy.com', '*!person@watching.us')
        self.assertEqual(self.events, self.cannedModes)

    def test_fjoin(self):
        self.protocol.irc_FJOIN('test.chatspike.net', '#test', '1133992411', ',Omster ,Brain')
        self.assertEqual(self.events,
                         [('joined', '#test', 'Omster'),
                          ('joined', '#test', 'Brain')])

    def test_join(self):
        self.protocol.irc_JOIN('TestNick', '#foo,#bar', '1133992411')
        self.assertEqual(self.events,
                         [('joined', '#foo', 'TestNick'),
                          ('joined', '#bar', 'TestNick')])

    def test_part(self):
        self.protocol.irc_PART('TestNick', '#foo')
        self.assertEqual(self.events, [('parted', '#foo', 'TestNick', '')])

    def test_partMessage(self):
        self.protocol.irc_PART('TestNick', '#foo', 'baz')
        self.assertEqual(self.events, [('parted', '#foo', 'TestNick', 'baz')])

    def test_getTargetKind(self):
        """
        Test that channels and users are differentiated correctly.
        """
        self.assertEqual(self.protocol.getTargetKind('#foo'), 'channel')
        self.assertEqual(self.protocol.getTargetKind('#'), 'channel')
        self.assertEqual(self.protocol.getTargetKind('bar'), 'user')
        self.assertRaises(IRCBadMessage, self.protocol.getTargetKind, '')
