from datetime import timedelta
from textwrap import dedent

from twisted.trial.unittest import TestCase

from kali.controller import Pseudo, PseudoAvatar
from kali.errors import InvalidExpiryTime
from kali.util import usage, tabulate, parseExpiryTime


class TestingController(object):
    def __init__(self):
        self.store = None


class TestingAvatar(PseudoAvatar):
    @usage('TEST <foo> [bar]')
    def cmd_test(self, pseudo, *params):
        """
        This is a test.

        Usage: TEST <foo> [bar]

        Seriously, this is only a test, nothing to see here.
        """

    def cmd_blank(self, pseudo, *params):
        pass


class TestingClient(Pseudo):
    anonymousAvatarFactory = TestingAvatar


class HelpTests(TestCase):
    def setUp(self):
        self.ctrl = TestingController()
        self.pseudo = TestingClient(self.ctrl)
        self.avatar = TestingAvatar(self.pseudo, None)

    def locateCommand(self, msg):
        params = msg.split()
        return self.pseudo.locateCommand(self.avatar, params)

    def test_defaultUsageFormatting(self):
        """
        Using default usage string.
        """
        cmd = self.locateCommand('blank')
        self.assertEqual(cmd.usage, 'No usage information')

    def test_usageFormatting(self):
        """
        Formatting a usage string.
        """
        cmd = self.locateCommand('test')
        self.assertEqual(cmd.usage, 'TEST <\002foo\002> [\002bar\002]')

    def test_defaultShortHelp(self):
        """
        Using default short-help string.
        """
        cmd = self.locateCommand('blank')
        self.assertEqual(cmd.shortHelp, 'No additional help.')

    def test_shortHelp(self):
        """
        Generating short-help string.
        """
        cmd = self.locateCommand('test')
        self.assertEqual(cmd.shortHelp, 'This is a test.')


class MiscTests(TestCase):
    def test_tabulateDefaultJoiner(self):
        """
        Test that tabulating data with the default joiner works as intended.
        """
        output = dedent("""\
        Col1    Col2  Col3
        1       test  k
        foobar  lol   zozlasdf""")

        headers = ['Col1', 'Col2', 'Col3']
        data = [['1', 'test', 'k'],
                ['foobar', 'lol', 'zozlasdf']]
        result = '\n'.join(tabulate(headers, data))
        self.assertEqual(result, output)

    def test_tabulateCustomJoiner(self):
        """
        Test that tabulating data with a custom joiner works as intended.
        """
        output = dedent("""\
        Col1   | Col2 | Col3
        1      | test | k
        foobar | lol  | zozlasdf""")

        headers = ['Col1', 'Col2', 'Col3']
        data = [['1', 'test', 'k'],
                ['foobar', 'lol', 'zozlasdf']]
        result = '\n'.join(tabulate(headers, data, joiner=' | '))
        self.assertEqual(result, output)

    def test_invalidExpiryTimes(self):
        """
        Test that attempting parse an invalid expiry time fails.
        """
        self.assertRaises(InvalidExpiryTime, parseExpiryTime, 'zozl')
        self.assertRaises(InvalidExpiryTime, parseExpiryTime, '+1x2x3x4x')
        self.assertRaises(InvalidExpiryTime, parseExpiryTime, '+0')

    def test_expiryTimes(self):
        """
        Test that parsing an expiry time results in the correct time data.
        """
        self.assertEqual(parseExpiryTime('+1d'), timedelta(days=1))
        self.assertEqual(parseExpiryTime('+1d2h'), timedelta(days=1, hours=2))
        self.assertEqual(parseExpiryTime('+1d2h3m'), timedelta(days=1, hours=2, minutes=3))
        self.assertEqual(parseExpiryTime('+1d2h3m4s'), timedelta(days=1, hours=2, minutes=3, seconds=4))
        self.assertEqual(parseExpiryTime('+1d4s'), timedelta(days=1, seconds=4))
        self.assertEqual(parseExpiryTime('+1d3m'), timedelta(days=1, minutes=3))
        self.assertEqual(parseExpiryTime('+1d3m4s'), timedelta(days=1, minutes=3, seconds=4))
