import re
from datetime import timedelta

from kali.ikali import ICommand
from kali.errors import UsageError, InvalidExpiryTime


paramPattern = re.compile(r'([<[])(\w+)([>\]])')

def formatUsage(s):
    return paramPattern.sub(r'\1\002\2\002\3', s)


def usage(desc, min=None, max=None):
    """
    Decorate a function with usage information.

    After decorating a function with C{usage} it gains 3 new attributes::

        C{usage}::
            A marked up usage string generated from C{desc}.

        C{help}::
            A help string copied and dedented from C{__doc__}.

        C{arglimits}::
            A tuple of the minimum and maximum arguments from C{min} and
            C{max}, or C{None}.

    @param desc: A description of the usage. Angle brackets C{(<>)} should be
                 used to denote mandatory arguments while square brackets C{([])}
                 should be used for optional arguments.
    @type desc: C{str}

    @param min: The minimum number of arguments to enforce or C{None}. C{None}
                means that the implementation should decide on a suitable
                minumum.
    @type min: C{int} or C{None}

    @param max: The maximum number of arguments to enforce or C{None}. C{None}
                means that the implementation should decide on a suitable
                maximum.
    @type min: C{int} or C{None}
    """
    def fact(f):
        f.usage = formatUsage(desc)
        f.help = f.__doc__
        f.arglimits = (min, max)
        return f
    return fact


def formatHelpOverview(commands):
    if len(commands) == 0:
        return

    mnl = max(len(c.name) for c in commands)
    for cmd in commands:
        name = cmd.name
        padding = ' ' * (mnl - len(name))
        yield '%s  %s%s' % (name, padding, cmd.shortHelp)


_hiddenHelpOverviews = ['cmd_help']

class HelpMixin(object):
    help = ''

    @property
    def commands(self):
        commands = [ICommand(getattr(self, attrName)) for attrName in dir(self)
                    if attrName.startswith('cmd_') and attrName not in _hiddenHelpOverviews]

        commands.sort(key=lambda x: x.name)
        return commands

    def noticelines(self, message):
        if isinstance(message, (str, unicode)):
            message = message.splitlines()

        for line in message:
            self.notice(line or ' ')

    def showHelpOverview(self, params):
        if params:
            try:
                cmd, junk = self.locateCommand([params[0]])
                self.notice('Usage: %s' % (cmd.usage,))
                self.notice(' ')
                self.noticelines(cmd.help)
            except UsageError, e:
                self.notice(e)
        else:
            self.noticelines(formatHelpOverview(self.commands))

    @usage('HELP [command]')
    def cmd_help(self, pseudo, *params):
        if len(params) == 0:
            self.noticelines(self.help)
            self.notice(' ')
        self.showHelpOverview(params)


def tabulate(headers, data, joiner='  '):
    """
    Tabulate data, attaching the specified headers.

    @param headers: The table headers
    @type headers: C{iterable}

    @param data: The table data to be split into columns under each respective
                 heading
    @type data: C{iterable} of C{iterable}s

    @param joiner: The string used to join columns together
    @type joiner: C{str}

    @returns: The tabulated data in rows
    @rtype: C{iterable} of C{str}
    """
    data = [headers] + list(data)
    columnWidths = [max(map(len, c)) for c in zip(*data)]

    for row in data:
        yield joiner.join(value.ljust(columnWidths[i]) for i, value in enumerate(row)).rstrip()


_expiryPattern = re.compile(r'\+(\d+[dD])?(\d+[hH])?(\d+[mM])?(\d+[sS])?')
_expiryParams = {
    'd': 'days',
    'h': 'hours',
    'm': 'minutes',
    's': 'seconds'}

def parseExpiryTime(s):
    """
    Parse an expiry time string into a C{timedelta}.

    Expiry times are of the format::
        C{+[#d][#h][#m][#s]}

    Where C{+} is literal and C{#} represents an integer.

    @param s: The expiry date string, conforming to the above syntax to parse
    @type s: C{str}

    @raises: C{InvalidExpiryTime} if the syntax of C{s} is invalid

    @returns: The specified time period as structured data
    @rtype: L{datetime.timedelta}
    """
    matches = _expiryPattern.search(s)
    if matches is None or matches.groups() == (None, None, None, None):
        raise InvalidExpiryTime('Expiry time syntax is: +[#d][#h][#m][#s] at least one group must be present')

    groups = matches.groups()
    params = dict((_expiryParams.get(v[-1]), int(v[:-1])) for v in groups if v is not None)
    return timedelta(**params)
